@@ -1,7 +1,7 @@
 /*
 
 asm6809, a Motorola 6809 cross assembler
-Copyright 2013-2014 Ciaran Anscomb
+Copyright 2013-2017 Ciaran Anscomb
 
 This program is free software: you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
@@ -18,9 +18,11 @@ option) any later version.
 #include <stdio.h>
 #include <stdlib.h>
 
-#include "c-strcase.h"
-
 #include "array.h"
+#include "c-strcase.h"
+#include "dict.h"
+#include "slist.h"
+
 #include "asm6809.h"
 #include "assemble.h"
 #include "error.h"
@@ -41,7 +43,13 @@ static int defining_macro_level = 0;
 static unsigned asm_pass;
 static unsigned prog_depth = 0;
 
-static void set_label(struct node *label, struct node *value);
+enum cond_state {
+	cond_state_if,
+	cond_state_if_done,
+	cond_state_else
+};
+
+static void set_label(struct node *label, struct node *value, _Bool changeable);
 static void args_float_to_int(struct node *args);
 static int verify_num_args(struct node *args, int min, int max, const char *op);
 static int64_t have_int_optional(struct node *args, int aindex, const char *op, int64_t in);
@@ -54,11 +62,14 @@ static void pseudo_endm(struct prog_line *);
 static void pseudo_export(struct prog_line *);
 
 static void pseudo_equ(struct prog_line *);
+static void pseudo_set(struct prog_line *);
 static void pseudo_org(struct prog_line *);
 static void pseudo_section(struct prog_line *);
 static void pseudo_section_name(struct prog_line *line);
 
 static void pseudo_fcc(struct prog_line *);
+static void pseudo_fcn(struct prog_line *);
+static void pseudo_fcs(struct prog_line *);
 static void pseudo_fdb(struct prog_line *);
 static void pseudo_rzb(struct prog_line *);
 static void pseudo_fill(struct prog_line *);
@@ -68,6 +79,7 @@ static void pseudo_put(struct prog_line *);
 static void pseudo_setdp(struct prog_line *);
 static void pseudo_include(struct prog_line *);
 static void pseudo_includebin(struct prog_line *);
+static void pseudo_end(struct prog_line *);
 static void pseudo_nop(struct prog_line *);
 
 struct pseudo_op {
@@ -77,8 +89,9 @@ struct pseudo_op {
 
 /* Pseudo-ops that override any label meaning */
 
-static struct pseudo_op label_ops[] = {
+static struct pseudo_op pseudo_label_ops[] = {
 	{ .name = "equ", .handler = &pseudo_equ },
+	{ .name = "set", .handler = &pseudo_set },
 	{ .name = "org", .handler = &pseudo_org },
 	{ .name = "section", .handler = &pseudo_section },
 	{ .name = "code", .handler = &pseudo_section_name },
@@ -93,8 +106,11 @@ static struct pseudo_op label_ops[] = {
 static struct pseudo_op pseudo_data_ops[] = {
 	{ .name = "fcc", .handler = &pseudo_fcc },
 	{ .name = "fcb", .handler = &pseudo_fcc },  // treat the same
+	{ .name = "fcn", .handler = &pseudo_fcn },
+	{ .name = "fcs", .handler = &pseudo_fcs },
 	{ .name = "fdb", .handler = &pseudo_fdb },
 	{ .name = "rzb", .handler = &pseudo_rzb },
+	{ .name = "fzb", .handler = &pseudo_rzb },
 	{ .name = "zmb", .handler = &pseudo_rzb },  // alias
 	{ .name = "bsz", .handler = &pseudo_rzb },  // alias
 	{ .name = "fill", .handler = &pseudo_fill },
@@ -108,6 +124,7 @@ static struct pseudo_op pseudo_ops[] = {
 	{ .name = "setdp", .handler = &pseudo_setdp },
 	{ .name = "include", .handler = &pseudo_include },
 	{ .name = "includebin", .handler = &pseudo_includebin },
+	{ .name = "end", .handler = &pseudo_end },
 	{ .name = "page", .handler = &pseudo_nop },
 	{ .name = "opt", .handler = &pseudo_nop },
 	{ .name = "spc", .handler = &pseudo_nop },
@@ -116,6 +133,50 @@ static struct pseudo_op pseudo_ops[] = {
 	{ .name = "name", .handler = &pseudo_nop },
 };
 
+/* Speed psuedo-op lookups using dictionaries */
+
+static struct dict *pseudo_label_dict = NULL;
+static struct dict *pseudo_data_dict = NULL;
+static struct dict *pseudo_dict = NULL;
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
+void assemble_init(void) {
+	if (!pseudo_label_dict) {
+		pseudo_label_dict = dict_new(dict_str_hash_case, dict_str_equal_case);
+		for (unsigned i = 0; i < ARRAY_N_ELEMENTS(pseudo_label_ops); i++) {
+			dict_insert(pseudo_label_dict, (void *)pseudo_label_ops[i].name, (void *)pseudo_label_ops[i].handler);
+		}
+	}
+	if (!pseudo_data_dict) {
+		pseudo_data_dict = dict_new(dict_str_hash_case, dict_str_equal_case);
+		for (unsigned i = 0; i < ARRAY_N_ELEMENTS(pseudo_data_ops); i++) {
+			dict_insert(pseudo_data_dict, (void *)pseudo_data_ops[i].name, (void *)pseudo_data_ops[i].handler);
+		}
+	}
+	if (!pseudo_dict) {
+		pseudo_dict = dict_new(dict_str_hash_case, dict_str_equal_case);
+		for (unsigned i = 0; i < ARRAY_N_ELEMENTS(pseudo_ops); i++) {
+			dict_insert(pseudo_dict, (void *)pseudo_ops[i].name, (void *)pseudo_ops[i].handler);
+		}
+	}
+}
+
+void assemble_free_all(void) {
+	if (pseudo_label_dict) {
+		dict_destroy(pseudo_label_dict);
+		pseudo_label_dict = NULL;
+	}
+	if (pseudo_data_dict) {
+		dict_destroy(pseudo_data_dict);
+		pseudo_data_dict = NULL;
+	}
+	if (pseudo_dict) {
+		dict_destroy(pseudo_dict);
+		pseudo_dict = NULL;
+	}
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
 static enum node_attr arg_attr(struct node *args, int index) {
@@ -202,6 +263,8 @@ static int64_t have_int_required(struct node *args, int aindex, const char *op,
 	return have_int_optional(args, aindex, op, in);
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
 /* Perform an assembly pass on a program. */
 
 void assemble_prog(struct prog *prog, unsigned pass) {
@@ -213,6 +276,12 @@ void assemble_prog(struct prog *prog, unsigned pass) {
 	prog_depth++;
 	struct prog_ctx *ctx = prog_ctx_new(prog);
 
+	/* cond_excluded will point to the element in cond_list that started to
+	 * exclude code.  ENDIF will stop excluding code if back to that
+	 * position. */
+	struct slist *cond_list = NULL;
+	struct slist *cond_excluded = NULL;
+
 	while (!prog_ctx_end(ctx)) {
 		/* Dummy line to be populated with values evaluated or not, as
 		 * appropriate. */
@@ -239,18 +308,18 @@ void assemble_prog(struct prog *prog, unsigned pass) {
 
 		/* Macro handling */
 
-		if (n_line.opcode && 0 == c_strcasecmp("macro", n_line.opcode->data.as_string)) {
+		if (!cond_excluded && n_line.opcode && 0 == c_strcasecmp("macro", n_line.opcode->data.as_string)) {
 			defining_macro_level++;
 			if (defining_macro_level == 1) {
 				n_line.label = eval_string(l->label);
-				n_line.args = eval_node(l->args);
+				n_line.args = node_ref(l->args);
 				pseudo_macro(&n_line);
 				listing_add_line(-1, 0, NULL, l->text);
 				goto next_line;
 			}
 		}
 
-		if (n_line.opcode && 0 == c_strcasecmp("endm", n_line.opcode->data.as_string)) {
+		if (!cond_excluded && n_line.opcode && 0 == c_strcasecmp("endm", n_line.opcode->data.as_string)) {
 			if (defining_macro_level == 0) {
 				error(error_type_syntax, "ENDM without beginning MACRO");
 				goto next_line;
@@ -271,6 +340,86 @@ void assemble_prog(struct prog *prog, unsigned pass) {
 			goto next_line;
 		}
 
+		/* Conditional assembly */
+
+		if (n_line.opcode && 0 == c_strcasecmp("if", n_line.opcode->data.as_string)) {
+			listing_add_line(-1, 0, NULL, l->text);
+			if (!cond_excluded) {
+				symbol_ignore_undefined = 1;
+				n_line.args = eval_node(l->args);
+				symbol_ignore_undefined = 0;
+				if (verify_num_args(n_line.args, 1, 1, "IF") < 0)
+					goto next_line;
+			}
+			if (!cond_excluded && !have_int_required(n_line.args, 0, "IF", 0)) {
+				cond_list = slist_prepend(cond_list, (void *)cond_state_if);
+				cond_excluded = cond_list;
+			} else {
+				cond_list = slist_prepend(cond_list, (void *)cond_state_if_done);
+			}
+			goto next_line;
+		}
+
+		if (n_line.opcode && 0 == c_strcasecmp("elsif", n_line.opcode->data.as_string)) {
+			listing_add_line(-1, 0, NULL, l->text);
+			if (!cond_list) {
+				error(error_type_syntax, "ELSIF without IF");
+			} else if ((intptr_t)cond_list->data == cond_state_else) {
+				error(error_type_syntax, "repeated ELSE");
+			} else {
+				if (!cond_excluded) {
+					cond_excluded = cond_list;
+				} else if (cond_excluded == cond_list &&
+					   (intptr_t)cond_list->data != cond_state_if_done) {
+					symbol_ignore_undefined = 1;
+					n_line.args = eval_node(l->args);
+					symbol_ignore_undefined = 0;
+					if (verify_num_args(n_line.args, 1, 1, "ELSIF") < 0)
+						goto next_line;
+					if (have_int_required(n_line.args, 0, "ELSIF", 0)) {
+						cond_excluded = NULL;
+						cond_list->data = (void *)cond_state_if_done;
+					}
+				}
+			}
+			goto next_line;
+		}
+
+		if (n_line.opcode && 0 == c_strcasecmp("else", n_line.opcode->data.as_string)) {
+			listing_add_line(-1, 0, NULL, l->text);
+			if (!cond_list) {
+				error(error_type_syntax, "ELSE without IF");
+			} else if ((intptr_t)cond_list->data == cond_state_else) {
+				error(error_type_syntax, "repeated ELSE");
+			} else if (cond_excluded == cond_list) {
+				if ((intptr_t)cond_list->data != cond_state_if_done) {
+					cond_list->data = (void *)cond_state_else;
+					cond_excluded = NULL;
+				}
+			} else if (!cond_excluded) {
+				cond_list->data = (void *)cond_state_else;
+				cond_excluded = cond_list;
+			}
+			goto next_line;
+		}
+
+		if (n_line.opcode && 0 == c_strcasecmp("endif", n_line.opcode->data.as_string)) {
+			listing_add_line(-1, 0, NULL, l->text);
+			if (!cond_list) {
+				error(error_type_syntax, "ENDIF without IF");
+				goto next_line;
+			} else if (cond_excluded == cond_list) {
+				cond_excluded = NULL;
+			}
+			cond_list = slist_remove(cond_list, cond_list->data);
+			goto next_line;
+		}
+
+		if (cond_excluded) {
+			listing_add_line(-1, 0, NULL, l->text);
+			goto next_line;
+		}
+
 		/* Normal processing */
 
 		n_line.label = eval_int(l->label);
@@ -289,18 +438,18 @@ void assemble_prog(struct prog *prog, unsigned pass) {
 		n_line.args = eval_node(l->args);
 
 		/* Pseudo-ops which determine a label's value */
+		void (*op_handler)(struct prog_line *);
 		if (n_line.opcode) {
-			for (unsigned i = 0; i < ARRAY_N_ELEMENTS(label_ops); i++) {
-				if (0 == c_strcasecmp(label_ops[i].name, n_line.opcode->data.as_string)) {
-					label_ops[i].handler(&n_line);
-					goto next_line;
-				}
+			op_handler = dict_lookup(pseudo_label_dict, n_line.opcode->data.as_string);
+			if (op_handler) {
+				op_handler(&n_line);
+				goto next_line;
 			}
 		}
 
 		/* Otherwise, any label on the line gets PC as its value */
 		if (n_line.label) {
-			set_label(n_line.label, node_new_int(cur_section->pc));
+			set_label(n_line.label, node_new_int(cur_section->pc), 0);
 		}
 
 		/* No opcode?  Next line. */
@@ -311,26 +460,24 @@ void assemble_prog(struct prog *prog, unsigned pass) {
 		}
 
 		/* Pseudo-ops that emit or reserve data */
-		for (unsigned i = 0; i < ARRAY_N_ELEMENTS(pseudo_data_ops); i++) {
-			if (0 == c_strcasecmp(pseudo_data_ops[i].name, n_line.opcode->data.as_string)) {
-				int old_pc = cur_section->pc;
-				pseudo_data_ops[i].handler(&n_line);
-				int nbytes = cur_section->pc - old_pc;
-				if (cur_section->span && cur_section->pc == (int)(cur_section->span->put + cur_section->span->size))
-					listing_add_line(old_pc & 0xffff, nbytes, cur_section->span, l->text);
-				else
-					listing_add_line(old_pc & 0xffff, nbytes, NULL, l->text);
-				goto next_line;
-			}
+		op_handler = dict_lookup(pseudo_data_dict, n_line.opcode->data.as_string);
+		if (op_handler) {
+			int old_pc = cur_section->pc;
+			op_handler(&n_line);
+			int nbytes = cur_section->pc - old_pc;
+			if (cur_section->span && cur_section->pc == (int)(cur_section->span->put + cur_section->span->size))
+				listing_add_line(old_pc & 0xffff, nbytes, cur_section->span, l->text);
+			else
+				listing_add_line(old_pc & 0xffff, nbytes, NULL, l->text);
+			goto next_line;
 		}
 
 		/* Other pseudo-ops */
-		for (unsigned i = 0; i < ARRAY_N_ELEMENTS(pseudo_ops); i++) {
-			if (0 == c_strcasecmp(pseudo_ops[i].name, n_line.opcode->data.as_string)) {
-				listing_add_line(-1, 0, NULL, l->text);
-				pseudo_ops[i].handler(&n_line);
-				goto next_line;
-			}
+		op_handler = dict_lookup(pseudo_dict, n_line.opcode->data.as_string);
+		if (op_handler) {
+			listing_add_line(-1, 0, NULL, l->text);
+			op_handler(&n_line);
+			goto next_line;
 		}
 
 		/* Real instructions */
@@ -393,6 +540,11 @@ next_line:
 		node_free(n_line.args);
 	}
 
+	if (cond_list) {
+		slist_free(cond_list);
+		error(error_type_syntax, "IF not matched with ENDIF");
+	}
+
 	assert(prog_depth > 0);
 	prog_depth--;
 	prog_ctx_free(ctx);
@@ -403,7 +555,7 @@ next_line:
 /* A disposable node must be passed in as value.  symbol_set() performs an eval
  * and stores the result, not the original node. */
 
-static void set_label(struct node *label, struct node *value) {
+static void set_label(struct node *label, struct node *value, _Bool changeable) {
 	switch (node_type_of(label)) {
 	default:
 		error(error_type_syntax, "invalid label type");
@@ -414,7 +566,7 @@ static void set_label(struct node *label, struct node *value) {
 		symbol_local_set(cur_section->local_labels, label->data.as_int, cur_section->line_number, value, asm_pass);
 		break;
 	case node_type_string:
-		symbol_set(label->data.as_string, value, asm_pass);
+		symbol_set(label->data.as_string, value, changeable, asm_pass);
 		break;
 	}
 	node_free(value);
@@ -430,7 +582,23 @@ static void pseudo_equ(struct prog_line *line) {
 	if (verify_num_args(line->args, 1, 1, "EQU") < 0)
 		return;
 	struct node **arga = node_array_of(line->args);
-	set_label(line->label, node_ref(arga[0]));
+	set_label(line->label, node_ref(arga[0]), 0);
+	struct node *n = eval_int(arga[0]);
+	if (n) {
+		listing_add_line(n->data.as_int & 0xffff, 0, NULL, line->text);
+		node_free(n);
+	} else {
+		listing_add_line(-1, 0, NULL, line->text);
+	}
+}
+
+/* SET.  As EQU, but symbol is permitted to change value. */
+
+static void pseudo_set(struct prog_line *line) {
+	if (verify_num_args(line->args, 1, 1, "SET") < 0)
+		return;
+	struct node **arga = node_array_of(line->args);
+	set_label(line->label, node_ref(arga[0]), 1);
 	struct node *n = eval_int(arga[0]);
 	if (n) {
 		listing_add_line(n->data.as_int & 0xffff, 0, NULL, line->text);
@@ -449,7 +617,7 @@ static void pseudo_org(struct prog_line *line) {
 	cur_section->pc = new_pc;
 	if (new_pc >= 0)
 		cur_section->put = new_pc;
-	set_label(line->label, node_new_int(new_pc));
+	set_label(line->label, node_new_int(new_pc), 0);
 	listing_add_line(new_pc & 0xffff, 0, NULL, line->text);
 }
 
@@ -468,6 +636,7 @@ static void pseudo_section(struct prog_line *line) {
 	}
 	section_set(n->data.as_string, asm_pass);
 	node_free(n);
+	listing_add_line(-1, 0, NULL, line->text);
 }
 
 /*
@@ -523,6 +692,37 @@ static void pseudo_export(struct prog_line *line) {
 	}
 }
 
+static void emit_formatted(char const *ins, int argn, struct node *n, _Bool basic_token) {
+	uint8_t or_last = basic_token ? 0x80 : 0;
+	switch (node_type_of(n)) {
+	default:
+		error(error_type_syntax,
+		      "argument %d of '%s' invalid: expected string or integer value",
+		      argn, ins);
+		break;
+	case node_type_undef:
+		section_emit_pad(1);
+		break;
+	case node_type_empty:
+		section_emit_uint8(0);
+		break;
+	case node_type_int:
+		section_emit_uint8(n->data.as_int | or_last);
+		break;
+	case node_type_float:
+		section_emit_uint8((int32_t)n->data.as_float | or_last);
+		break;
+	case node_type_string:
+		for (char const *c = n->data.as_string; *c; c++) {
+			uint8_t cc = *c;
+			if (!*(c+1))
+				cc |= or_last;
+			section_emit_uint8(cc);
+		}
+		break;
+	}
+}
+
 /* FCC, FCB.  Embed string and byte constants. */
 
 static void pseudo_fcc(struct prog_line *line) {
@@ -531,31 +731,32 @@ static void pseudo_fcc(struct prog_line *line) {
 		return;
 	struct node **arga = node_array_of(line->args);
 	for (int i = 0; i < nargs; i++) {
-		struct node *arg = arga[i];
-		switch (node_type_of(arg)) {
-		default:
-			error(error_type_syntax,
-			      "argument %d of 'FCC' invalid: expected string or integer value",
-			      i+1);
-			break;
-		case node_type_undef:
-			section_emit_pad(1);
-			break;
-		case node_type_empty:
-			section_emit_uint8(0);
-			break;
-		case node_type_int:
-			section_emit_uint8(arg->data.as_int);
-			break;
-		case node_type_float:
-			section_emit_uint8((int32_t)arg->data.as_float);
-			break;
-		case node_type_string:
-			for (char const *c = arg->data.as_string; *c; c++) {
-				section_emit_uint8(*c);
-			}
-			break;
-		}
+		emit_formatted("FCC", i, arga[i], 0);
+	}
+}
+
+/* FCN.  As FCC, but zero-terminate. */
+
+static void pseudo_fcn(struct prog_line *line) {
+	int nargs = verify_num_args(line->args, 1, -1, "FCN");
+	if (nargs < 0)
+		return;
+	struct node **arga = node_array_of(line->args);
+	for (int i = 0; i < nargs; i++) {
+		emit_formatted("FCN", i, arga[i], 0);
+	}
+	section_emit_uint8(0);
+}
+
+/* FCS.  As FCC, but last byte gets top bit set. */
+
+static void pseudo_fcs(struct prog_line *line) {
+	int nargs = verify_num_args(line->args, 1, -1, "FCS");
+	if (nargs < 0)
+		return;
+	struct node **arga = node_array_of(line->args);
+	for (int i = 0; i < nargs; i++) {
+		emit_formatted("FCS", i, arga[i], 1);
 	}
 }
 
@@ -626,10 +827,12 @@ static void pseudo_include(struct prog_line *line) {
 		error(error_type_syntax, "invalid argument to INCLUDE");
 		return;
 	}
+	struct section *old_section = cur_section;
 	struct prog *file = prog_new_file(arga[0]->data.as_string);
 	if (!file)
 		return;
 	assemble_prog(file, asm_pass);
+	cur_section = old_section;
 }
 
 /* INCLUDEBIN.  Include a binary object in-place.  Unlike INCLUDE, the filename
@@ -656,20 +859,17 @@ static void pseudo_includebin(struct prog_line *line) {
 	fclose(f);
 }
 
-/* MACRO.  Start defining a named macro.  The macro name can either be
- * specified as an argument or as the label for the line the directive appears
- * on. */
+/* MACRO.  Start defining a named macro.  The line's label field is used as the
+ * macro name.  Arguments are ignored. */
 
 static void pseudo_macro(struct prog_line *line) {
-	int nargs = node_array_count(line->args);
-	struct node **arga = node_array_of(line->args);
 	const char *name;
-	if (nargs == 1 && !line->label) {
-		name = arga[0]->data.as_string;
-	} else if (nargs == 0 && node_type_of(line->label) == node_type_string) {
+	switch (node_type_of(line->label)) {
+	case node_type_string:
 		name = line->label->data.as_string;
-	} else {
-		error(error_type_syntax, "macro name must either be label OR argument");
+		break;
+	default:
+		error(error_type_syntax, "missing or invalid macro name");
 		return;
 	}
 	struct prog *macro = prog_macro_by_name(name);
@@ -705,6 +905,16 @@ static void pseudo_endm(struct prog_line *line) {
 	defining_macro_ctx = NULL;
 }
 
+/* END.  Specify EXEC address. */
+
+static void pseudo_end(struct prog_line *line) {
+	int nargs = verify_num_args(line->args, 0, 1, "END");
+	if (nargs < 1)
+		return;
+	struct node **arga = node_array_of(line->args);
+	symbol_set(".exec", arga[0], asm_pass, 0);
+}
+
 /* Ignore certain historical pseudo-ops */
 
 static void pseudo_nop(struct prog_line *line) {
