@@ -1,7 +1,7 @@
 /*
 
 asm6809, a Motorola 6809 cross assembler
-Copyright 2013-2014 Ciaran Anscomb
+Copyright 2013-2017 Ciaran Anscomb
 
 This program is free software: you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
@@ -16,10 +16,10 @@ option) any later version.
 struct prog;
 
 /*
- * Assemble a file or macro.
- * */
+ * Required initialisation.
+ */
 
-void assemble_prog(struct prog *file, unsigned pass);
+void assemble_init(void);
 
 /*
  * Tidy up.
@@ -27,4 +27,10 @@ void assemble_prog(struct prog *file, unsigned pass);
 
 void assemble_free_all(void);
 
-#endif  /* ASM6809_ASSEMBLE_H_ */
+/*
+ * Assemble a file or macro.
+ */
+
+void assemble_prog(struct prog *file, unsigned pass);
+
+#endif
