@@ -1,7 +1,7 @@
 /*
 
 asm6809, a Motorola 6809 cross assembler
-Copyright 2013-2014 Ciaran Anscomb
+Copyright 2013-2017 Ciaran Anscomb
 
 This program is free software: you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
@@ -16,6 +16,7 @@ option) any later version.
 #include <stdlib.h>
 
 #include "array.h"
+#include "asm6809.h"
 #include "assemble.h"
 #include "error.h"
 #include "eval.h"
@@ -53,21 +54,18 @@ void instr_immediate(struct opcode const *op, struct node const *args) {
 		return;
 	}
 	section_emit_op(op->immediate);
-	if (node_type_of(arga[0]) == node_type_int) {
-		if ((op->type & OPCODE_EXT_TYPE) == OPCODE_IMM8)
-			section_emit_uint8(arga[0]->data.as_int);
-		else if ((op->type & OPCODE_EXT_TYPE) == OPCODE_IMM16)
-			section_emit_uint16(arga[0]->data.as_int);
-		else
-			section_emit_uint32(arga[0]->data.as_int);
-	} else {
-		if ((op->type & OPCODE_EXT_TYPE) == OPCODE_IMM8)
-			section_emit_pad(1);
-		else if ((op->type & OPCODE_EXT_TYPE) == OPCODE_IMM16)
-			section_emit_pad(2);
-		else
-			section_emit_pad(4);
+	struct node *tmp = eval_int(arga[0]);
+	uint32_t value = 0;
+	if (tmp) {
+		value = tmp->data.as_int;
+		node_free(tmp);
 	}
+	if ((op->type & OPCODE_EXT_TYPE) == OPCODE_IMM8)
+		section_emit_uint8(value);
+	else if ((op->type & OPCODE_EXT_TYPE) == OPCODE_IMM16)
+		section_emit_uint16(value);
+	else
+		section_emit_uint32(value);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -184,7 +182,7 @@ static struct {
 
 static _Bool off_type_compatible(enum off_type off_type, _Bool pcr, struct node const *n) {
 	enum node_type ntype = node_type_of(n);
-	enum node_type nattr = node_attr_of(n);
+	enum node_attr nattr = node_attr_of(n);
 
 	switch (nattr) {
 	case node_attr_none:
@@ -299,7 +297,7 @@ static void instr_indexed2(_Bool indirect, struct node const *arg0, struct node
 	enum off_type off_type;
 	enum idx_indirect idx_indirect;
 	int postbyte = -1;
-	for (int i = 0; i < NUM_INDEXED_MODES; i++) {
+	for (unsigned i = 0; i < NUM_INDEXED_MODES; i++) {
 		if (indexed_modes[i].idx_type != idx_type)
 			continue;
 		if (indexed_modes[i].idx_attr != arg1_attr)
@@ -368,7 +366,7 @@ void instr_indexed(struct opcode const *op, struct node const *args, int imm8_va
 	int nargs = node_array_count(args);
 	struct node **arga = node_array_of(args);
 	int indirect = 0;
-	if (nargs == 1 && arga[0]->type == node_type_array) {
+	if (nargs == 1 && node_type_of(arga[0]) == node_type_array) {
 		indirect = 0x10;
 		nargs = node_array_count(arga[0]);
 		arga = arga[0]->data.as_array.args;
@@ -381,6 +379,12 @@ void instr_indexed(struct opcode const *op, struct node const *args, int imm8_va
 	if (imm8_val >= 0)
 		section_emit_uint8(imm8_val);
 
+	if (node_type_of(arga[0]) == node_type_string) {
+		struct node *tmp = eval_int(arga[0]);
+		node_free(arga[0]);
+		arga[0] = tmp;
+	}
+
 	if (nargs == 1) {
 		int pbyte = indirect ? 0x9f : 0x8f;
 		int addr = 0;
@@ -396,6 +400,7 @@ void instr_indexed(struct opcode const *op, struct node const *args, int imm8_va
 			return;
 		}
 		switch (node_attr_of(arga[0])) {
+		case node_attr_undef:
 		case node_attr_none:
 		case node_attr_16bit:
 			break;
@@ -540,13 +545,18 @@ void instr_pair(struct opcode const *op, struct node const *args) {
 			size[i] = reg_tfr_size[arga[i]->data.as_reg];
 		} else if (type[i] == node_type_int) {
 			int v = arga[i]->data.as_int;
-			error(error_type_illegal, "numerical values used in inter-register op");
-			if (v < 0 || v > 15) {
-				error(error_type_out_of_range, "numeric value for register out of range");
+			if (v == 0 && asm6809_options.isa == asm6809_isa_6309) {
+				nibble[i] = 0x0c;
+				size[i] = 16 | 8;
+			} else {
+				error(error_type_illegal, "numerical values used in inter-register op");
+				if (v < 0 || v > 15) {
+					error(error_type_out_of_range, "numeric value for register out of range");
+				}
+				nibble[i] = v & 15;
+				// don't warn about size as well if numeric...
+				size[i] = 16 | 8;
 			}
-			nibble[i] = v & 15;
-			// don't warn about size as well if numeric...
-			size[i] = 16 | 8;
 		} else if (type[i] != node_type_undef) {
 			error(error_type_syntax, "invalid register in inter-register op");
 			return;
@@ -560,10 +570,6 @@ void instr_pair(struct opcode const *op, struct node const *args) {
 	section_emit_op(op->immediate);
 	section_emit_uint8(pbyte);
 	return;
-
-invalid_register:
-	error(error_type_syntax, "invalid register in inter-register op");
-	return;
 }
 
 /*
@@ -586,8 +592,10 @@ void instr_tfm(struct opcode const *op, struct node const *args) {
 		attr[i] = node_attr_of(arga[i]);
 		if (type[i] == node_type_reg) {
 			int v = reg_tfr_nibble[arga[i]->data.as_reg];
-			if (v < 0 || v > 4)
-				goto invalid_register;
+			if (v < 0 || v > 4) {
+				error(error_type_syntax, "invalid register in TFM op");
+				return;
+			}
 			nibble[i] = v;
 		} else if (type[i] == node_type_int) {
 			int v = arga[i]->data.as_int;
@@ -620,10 +628,6 @@ void instr_tfm(struct opcode const *op, struct node const *args) {
 	section_emit_op(op->immediate + mod);
 	section_emit_uint8(pbyte);
 	return;
-
-invalid_register:
-	error(error_type_syntax, "invalid register in TFM op");
-	return;
 }
 
 /*
